package org.jboss.cache.config.parsing;

import org.jboss.cache.config.*;
import org.jboss.cache.config.parsing.custominterceptors.AaaCustomInterceptor;
import org.jboss.cache.config.parsing.custominterceptors.BbbCustomInterceptor;
import org.jboss.cache.eviction.LRUAlgorithmConfig;
import org.jboss.cache.eviction.MRUAlgorithmConfig;
import org.jboss.cache.lock.IsolationLevel;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;

/**
 * Parses a 'normal' configuration file and makes sure that the resulted <b>Configuration</b> object has
 * the expected state.
 *
 * @author Mircea.Markus@jboss.com
 * @since 3.0
 */
@Test(groups = "functional")
public class XmlConfigurationParserTest
{
   Configuration config;

   @BeforeMethod
   public void setUp()
   {
      XmlConfigurationParser parser = new XmlConfigurationParser(false, null);
      config = parser.parseFile("configs/parser-test.xml");
   }

   public void testParseOldConfigFile()
   {
      System.setProperty("jbosscache.config.validate","false");
      XmlConfigurationParser parser = new XmlConfigurationParser();
      try
      {
         parser.parseFile("configs/conf2x/pess-local.xml");
         assert false : "exception expected";
      }
      catch (ConfigurationException e)
      {
         //expectd
      }
      finally {
         System.setProperty("jbosscache.config.validate","true");
      }
   }

   public void testTransactionManagerLookupClass()
   {
      assert config.getTransactionManagerLookupClass().equals("org.jboss.cache.transaction.GenericTransactionManagerLookup");
   }

   public void testIsolationLevel()
   {
      assert config.getIsolationLevel().equals(IsolationLevel.REPEATABLE_READ);
   }

   public void testCacheMode()
   {
      assert config.getCacheMode().equals(Configuration.CacheMode.REPL_SYNC);
   }

   public void testUseReplQueue()
   {
      assert !config.isUseReplQueue();
   }

   public void testClusterName()
   {
      assert config.getClusterName().equals("JBossCache-Cluster");
   }

   public void testGetClusterConfig()
   {
      assert config.getClusterConfig().indexOf("MERGE2") >= 0;
   }

   public void testFetchInMemoryState()
   {
      assert config.isFetchInMemoryState();
   }

   public void testStateRetrievalTimeout()
   {
      assert config.getStateRetrievalTimeout() == 15124;
   }

   public void testSyncReplTimeout()
   {
      assert config.getSyncReplTimeout() == 15421;
   }

   public void testLockAcquisitionTimeout()
   {
      assert config.getLockAcquisitionTimeout() == 10234;
   }

   public void testUseLazyDeserialization()
   {
      assert config.isUseLazyDeserialization();
   }

   public void testObjectInputStreamPoolSize()
   {
      assert 12 == config.getObjectInputStreamPoolSize();
   }

   public void testObjectOutputStreamPoolSize()
   {
      assert 14 == config.getObjectOutputStreamPoolSize();
   }

   public void testShutdownHookBehavior()
   {
      assert Configuration.ShutdownHookBehavior.REGISTER == config.getShutdownHookBehavior();
   }

   public void testSyncRollbackPhase()
   {
      assert config.isSyncRollbackPhase();
   }

   public void testSyncCommitPhase()
   {
      assert config.isSyncCommitPhase();
   }

   public void testUseReplicationVersion()
   {
      assert config.getReplicationVersion() == 124;
   }

   public void testGetMultiplexerStack()
   {
      assert "file_name".equals(config.getMultiplexerStack());
   }

   public void testMarshallerClass()
   {
      assert "some.Clazz".equals(config.getMarshallerClass());
   }

   public void testLockParentForChildInsertRemove()
   {
      assert config.isLockParentForChildInsertRemove();
   }

   public void testInactiveOnStartup()
   {
      assert config.isInactiveOnStartup();
   }

   public void testExposeManagementStatistics()
   {
      assert !config.getExposeManagementStatistics();
   }

   public void testCacheLoaderConfiguration()
   {
      CacheLoaderConfig clc = config.getCacheLoaderConfig();
      assert null != clc;
      assert clc.isPassivation();
      assert clc.isShared();
      assert "/a/b/c,/f/r/s".equals(clc.getPreload());
      CacheLoaderConfig.IndividualCacheLoaderConfig first = clc.getFirstCacheLoaderConfig();
      assert "org.jboss.cache.loader.JDBCCacheLoader".equals(first.getClassName());
      assert first.isAsync();
      assert first.isFetchPersistentState();
      assert first.isIgnoreModifications();
      assert first.isPurgeOnStartup();
      assert first.getProperties().get("cache.jdbc.table.name").equals("jbosscache");
      assert first.getProperties().get("cache.jdbc.table.create").equals("true");
      assert first.getProperties().get("cache.jdbc.table.drop").equals("true");
   }

   public void testBuddyReplicationConfig()
   {
      BuddyReplicationConfig brConfig = config.getBuddyReplicationConfig();
      assert brConfig.isEnabled();
      BuddyReplicationConfig.BuddyLocatorConfig locatorConfig = brConfig.getBuddyLocatorConfig();
      assert "org.jboss.cache.buddyreplication.NextMemberBuddyLocator".equals(locatorConfig.getBuddyLocatorClass());
      assert locatorConfig.getBuddyLocatorProperties().get("numBuddies").equals("1");
      assert locatorConfig.getBuddyLocatorProperties().get("ignoreColocatedBuddies").equals("true");
      assert brConfig.getBuddyPoolName().equals("myBuddyPoolReplicationGroup");
      assert brConfig.getBuddyCommunicationTimeout() == 2000;
      assert brConfig.isAutoDataGravitation();
      assert brConfig.isDataGravitationRemoveOnFind();
      assert brConfig.isDataGravitationSearchBackupTrees();
   }

   public void testUseRegionBasedMarshalling()
   {
      assert config.isUseRegionBasedMarshalling();
   }

   public void testEvictionPolicyConfig()
   {
      EvictionConfig evictionConfig = config.getEvictionConfig();
      assert "org.jboss.cache.eviction.LRUAlgorithm".equals(evictionConfig.getDefaultEvictionRegionConfig().getEvictionAlgorithmConfig().getEvictionAlgorithmClassName());
      assert 200000 == evictionConfig.getDefaultEvictionRegionConfig().getEventQueueSize();
      assert 5 == evictionConfig.getWakeupInterval();

      List<EvictionRegionConfig> regionConfigs = evictionConfig.getEvictionRegionConfigs();
      assert regionConfigs.size() == 2;

      EvictionRegionConfig first = evictionConfig.getDefaultEvictionRegionConfig();
      assert first.getRegionName().equals("/");
      assert first.getEvictionAlgorithmConfig() instanceof LRUAlgorithmConfig;
      LRUAlgorithmConfig firstConfiguration = (LRUAlgorithmConfig) first.getEvictionAlgorithmConfig();
      assert firstConfiguration.getMaxAge() <= 0;
      assert firstConfiguration.getTimeToLive() == 1000;
      assert firstConfiguration.getMaxNodes() == 5000;

      EvictionRegionConfig second = regionConfigs.get(0);
      LRUAlgorithmConfig secondConfiguration = (LRUAlgorithmConfig) second.getEvictionAlgorithmConfig();
      assert secondConfiguration.getMaxAge() == -1;
      assert secondConfiguration.getTimeToLive() == 1002;
      assert secondConfiguration.getMaxNodes() == 0;

      EvictionRegionConfig third = regionConfigs.get(1);
      MRUAlgorithmConfig thirdConfiguration = (MRUAlgorithmConfig) third.getEvictionAlgorithmConfig();
      assert thirdConfiguration.getMaxNodes() == 2103;
      assert thirdConfiguration.getMinTimeToLive() == 22;
      assert third.getEventQueueSize() == 21;
   }

   public void testCustomInterceptors()
   {
      List<CustomInterceptorConfig> interceptorConfigs = config.getCustomInterceptors();
      assert interceptorConfigs.size() == 5;
      assert interceptorConfigs.get(0).getInterceptor() instanceof AaaCustomInterceptor;
      assert interceptorConfigs.get(1).getInterceptor() instanceof BbbCustomInterceptor;
      assert interceptorConfigs.get(2).getInterceptor() instanceof AaaCustomInterceptor;
      assert interceptorConfigs.get(3).getInterceptor() instanceof BbbCustomInterceptor;
      assert interceptorConfigs.get(4).getInterceptor() instanceof AaaCustomInterceptor;
      assert interceptorConfigs.get(0).isFirst();
      assert !interceptorConfigs.get(0).isLast();
      assert interceptorConfigs.get(1).isLast();
      assert interceptorConfigs.get(2).getIndex() == 3;
      assert interceptorConfigs.get(3).getBeforeClass().equals("org.jboss.cache.interceptors.CallInterceptor");
      assert interceptorConfigs.get(4).getAfterClass().equals("org.jboss.cache.interceptors.CallInterceptor");
   }

   public void testSingletonStore()
   {
      CacheLoaderConfig.IndividualCacheLoaderConfig clc = config.getCacheLoaderConfig().getFirstCacheLoaderConfig();
      assert clc != null;
      CacheLoaderConfig.IndividualCacheLoaderConfig.SingletonStoreConfig singlStoreConf = clc.getSingletonStoreConfig();
      assert singlStoreConf != null;
      assert singlStoreConf.isSingletonStoreEnabled();
      assert singlStoreConf.getSingletonStoreClass().equals("org.jboss.cache.loader.SingletonStoreCacheLoader");
      assert singlStoreConf.getProperties().size() == 2;
      assert singlStoreConf.getProperties().get("pushStateWhenCoordinator").equals("true");
      assert singlStoreConf.getProperties().get("pushStateWhenCoordinatorTimeout").equals("20000");
   }

   public void testMvccAttributes()
   {
      assert !config.isWriteSkewCheck();
      assert config.getConcurrencyLevel() == 21;
   }

   public void testListenerAsyncThreads()
   {
      assert config.getListenerAsyncPoolSize() == 5;
   }

   public void testInvocationBatching()
   {
      assert config.isInvocationBatchingEnabled();
   }
}
